from typing import Any, Callable, Optional, Tuple, Generator
from functools import reduce


def iterate_over(
    array: Any,
    method: Callable[
        [int, Tuple[int, ...], Optional[Tuple[int, ...]]], Tuple[int, ...]
    ],
    order: Optional[Tuple[int, ...]] = None,
) -> Generator[Any, None, None]:
    """
    Use a walker method to iterate over an array. The array can either be a
    plain python array or a numpy style array.
    """
    shape = []
    allows_multidimension = False

    # Handle numpy style arrays
    if hasattr(array, "shape"):
        shape = array.shape
        allows_multidimension = True
    else:
        # Fallback for nested lists or other iterable structures
        try:
            inside = array
            while True:
                shape.append(len(inside))
                inside = inside[0]
        except TypeError:
            pass

    shape = tuple(shape)

    size = reduce(lambda x, y: x * y, shape, 1)
    if allows_multidimension:
        # Quickly index multi-dimensional arrays
        for i in range(size):
            pos = method(i, shape, order)
            yield array[pos]
    else:

        def get(a, p):
            """Recursive function to get the value at a specific position."""
            if len(p) == 1:
                return a[p[0]]
            else:
                return get(a[p[0]], p[1:])

        for i in range(size):
            pos = method(i, shape, order)
            yield get(array, pos)


def flatten(
    array: Any,
    method: Callable[
        [int, Tuple[int, ...], Optional[Tuple[int, ...]]], Tuple[int, ...]
    ],
    order: Optional[Tuple[int, ...]] = None,
) -> Any:
    """
    Flattens a multi-dimensional array into a 1D array using the specified method.
    """
    return list(iterate_over(array, method, order))



from typing import Optional, Tuple


def lawnmower(
    index: int, shape: Tuple[int, ...], order: Optional[Tuple[int, ...]] = None
) -> Tuple[int, ...]:
    """
    Returns a position in the lawnmower pattern, given a shape to traverse
    over. If order is specified, perform the iteration over a different
    ordering of the axes.
    """

    # Setup an ascending order
    if order is None:
        order = tuple(range(len(shape)))

    # Check the size of the input are correct
    if len(order) != len(shape):
        raise ValueError("Order must match the number of dimensions in shape.")

    reshape = [shape[i] for i in order]

    # Convert flat index to normal coordinates (row-major order)
    coords = []
    remaining = index
    for dim in reversed(reshape):
        coords.append(remaining % dim)
        remaining //= dim
    coords.reverse()

    # Apply lawnmower flipping: flip axis i if the sum of parities of all previous axes is odd
    for i in range(1, len(coords)):
        flip = sum(coords[j] % 2 for j in range(i)) % 2 == 1
        if flip:
            coords[i] = reshape[i] - 1 - coords[i]

    recoords = [0] * len(shape)
    for pos, value in zip(order, coords):
        recoords[pos] = value
    return tuple(recoords)


def inverse_lawnmower(
    position: Tuple[int, ...],
    shape: Tuple[int, ...],
    order: Optional[Tuple[int, ...]] = None,
) -> int:
    """
    Returns the inverse of the lawnmower function, given a position in the
    walk, returns the index along the walk.
    """
    if order is None:
        order = tuple(range(len(shape)))

    if len(order) != len(shape):
        raise ValueError("Order must match the number of dimensions in shape.")

    # Reorder the position and shape
    reshape = [shape[i] for i in order]
    coords = [position[i] for i in order]

    # reverse the flipping that `lawnmower` applied
    for i in range(len(coords) - 1, 0, -1):
        # Compute whether a flip was applied at this axis
        flip = sum(coords[j] % 2 for j in range(i)) % 2 == 1
        if flip:
            coords[i] = reshape[i] - 1 - coords[i]

    # flatten coords to linear index (row-major)
    index = 0
    for i in range(len(coords)):
        index = index * reshape[i] + coords[i]

    return index



from typing import Any, Generator, Optional, Tuple


def zorder(
    index: int, shape: Tuple[int, ...], order: Optional[Tuple[int, ...]] = None
) -> Tuple[int, ...]:
    """
    Returns a position in the z order pattern, given a shape to traverse
    over. If order is specified, perform the iteration over a different
    ordering of the axes.
    """

    # Setup an ascending order
    if order is None:
        order = tuple(range(len(shape)))

    # Check the size of the input are correct
    if len(order) != len(shape):
        raise ValueError("Order must match the number of dimensions in shape.")

    if not all(s > 0 for s in shape):
        raise ValueError("All dimensions in shape must be greater than zero.")

    if not all(s == shape[0] for s in shape):
        raise ValueError("All dimensions in shape must be equal for z-ordering.")

    dim = shape[0]
    if (dim & (dim - 1)) != 0 and dim > 1:
        raise ValueError("Shape size must be a power of 2 for z-ordering.")

    bit_length = dim.bit_length()

    position = [0] * len(shape)

    for bit in range(bit_length):
        for axis in order:
            position[axis] |= (index & 0x1) << bit
            index >>= 1

    return tuple(position)


def zordered(
    array: Any, order: Optional[Tuple[int, ...]] = None
) -> Generator[Any, None, None]:
    """
    Returns a generator that yields elements of the array in z-order.
    If order is specified, perform the iteration over a different ordering
    of the axes.
    """
    if order is None:
        order = tuple(range(len(array.shape)))

    shape = array.shape

    if len(order) != len(shape):
        raise ValueError("Order must match the number of dimensions in array.")

    if not all(s > 0 for s in shape):
        raise ValueError("All dimensions in shape must be greater than zero.")

    dim = shape[0]
    if (dim & (dim - 1)) != 0 and dim > 1:
        raise ValueError("Shape size must be a power of 2 for z-ordering.")

    bit_length = dim.bit_length()

    for index in range(dim ** len(shape)):
        position = [0] * len(shape)
        for bit in range(bit_length):
            for axis in order:
                position[axis] |= (index & 0x1) << bit
                index >>= 1
        yield array[*position]


def inverse_zorder(
    position: Tuple[int, ...],
    shape: Tuple[int, ...],
    order: Optional[Tuple[int, ...]] = None,
) -> int:
    """
    Returns the inverse of the zorder function, given a position in the shape,
    return the index along the walk.
    """

    # Setup an ascending order
    if order is None:
        order = tuple(range(len(shape)))

    # Check the size of the input are correct
    if len(order) != len(shape) or len(position) != len(shape):
        raise ValueError("Order must match the number of dimensions in shape.")

    if not all(s > 0 for s in shape):
        raise ValueError("All dimensions in shape must be greater than zero.")

    if not all(s == shape[0] for s in shape):
        raise ValueError("All dimensions in shape must be equal for z-ordering.")

    dim = shape[0]
    if (dim & (dim - 1)) != 0 and dim > 1:
        raise ValueError("Shape size must be a power of 2 for z-ordering.")

    bit_length = dim.bit_length()

    index = 0
    for bit in reversed(range(bit_length)):
        for axis in reversed(order):
            bit_value = (position[axis] >> bit) & 0x1
            index = (index << 1) | bit_value

    return index



from typing import Tuple, Iterable, Generator, Dict


def hilbert(
    index: int,
    shape: Tuple[int, ...],
) -> Tuple[int, ...]:
    """
    Returns a position in the hilbert curve, given an index and a shape to follow.
    """

    dim = shape[0]
    for shape_dim in shape:
        if shape_dim <= 0:
            raise ValueError("Shape dimensions must be greater than zero.")
        if shape_dim != dim:
            raise ValueError("All dimensions in shape must be equal for Hilbert curve.")

    if (dim & (dim - 1)) != 0:
        raise ValueError("Shape dimension must be a power of two for Hilbert curve.")

    iterations = dim.bit_length() - 1

    if len(shape) == 2:

        def grow_2d_hilbert_path(path: Iterable[str]) -> Generator[str, None, None]:
            """Grow the Hilbert path by one iteration."""
            for char in path:
                if char == "A":
                    yield from "+BF-AFA-FB+"
                elif char == "B":
                    yield from "-AF+BFB+FA-"
                else:
                    yield char

        path = "A"
        for _ in range(iterations):
            path = grow_2d_hilbert_path(path)

        pos = (0, 0)
        dir = (0, 1)

        path = list(path)

        def iterate_2d(
            position: Tuple[int, int], direction: Tuple[int, int], next: str
        ) -> Tuple[Tuple[int, int], Tuple[int, int], bool]:
            """Iterate over hilbert path in 2d."""
            right_rotations: Dict[Tuple[int, int], Tuple[int, int]] = {
                (1, 0): (0, 1),
                (0, 1): (-1, 0),
                (-1, 0): (0, -1),
                (0, -1): (1, 0),
            }
            left_rotations = {b: a for a, b in right_rotations.items()}

            moved = False

            match next:
                case "F":
                    position = (position[0] + direction[0], position[1] + direction[1])
                    moved = True
                case "+":
                    direction = left_rotations[direction]
                case "-":
                    direction = right_rotations[direction]
                case "A" | "B":
                    pass

            return position, direction, moved

        movements = 0
        for char in path:
            if movements == index:
                break
            pos, dir, moved = iterate_2d(pos, dir, char)
            if moved:
                movements += 1

    return pos
